part of masamune.flutter;

/// Class that manages values.
/// 
/// Please use in conjunction with UIWidget.
/// 
/// Usually used to get the value using [get()] and [text()].
class UIValue {
  /// Root observer.
  /// 
  /// ```
  /// return new MaterialApp(
  ///   ...
  ///   navigatorObservers: <NavigatorObserver>[UIValue.routeObserver],
  ///   ...
  /// );
  /// ```
  static RouteObserver<PageRoute> get routeObserver => _routeObserver;
  static RouteObserver<PageRoute> _routeObserver = RouteObserver<PageRoute>();
  _UIWidgetState _state;
  /// List of Paths to dispose with when UIValue is disposed.
  PathList get willDisposePathList => this._willDisposePathList;
  PathList _willDisposePathList = PathList();
  Map<String, Observer> _data = MapPool.get();
  UIValue._( _UIWidgetState state ) : this._state = state;
  _UIWidgetContainerState _container;
  IDataDocument _form;
  int _updatedTime = 0;
  
  /// The closest instance of this class that encloses the given context.
  ///
  /// Typical usage:
  ///
  /// ```dart
  /// UIValue controller = UIValue.of(context);
  /// ```
  /// 
  /// [context]: Build context.
  static UIValue of(
    BuildContext context
  ) {
    final _UIWidgetScope scope
      = context.getElementForInheritedWidgetOfExactType<_UIWidgetScope>()?.widget;
    return scope?.target?._value;
  }
  /// Listen for the path in the UIWidget.
  /// 
  /// Please use it for update processing in the widget.
  /// 
  /// [path]: The path to get.
  void listen(
    String path
  ){
    path = path?.applyTags();
    if( isEmpty(path) ) return;
    this._addPath(path);
  }
  /// Get object from pathmap in conjunction with UIWidget.
  /// 
  /// You can get any IPath object other than String.
  /// 
  /// Please use it for update processing in the widget.
  /// 
  /// [path]: The path to get.
  /// [defaultValue]: Value if none.
  /// [filter]: Filter values.
  /// [listen]: True to listen for values.
  T get<T extends Object>(
    String path,
    {
      T defaultValue,
      T filter( T value ),
      bool listen = false
    }
  ){
    path = path?.applyTags();
    if( isEmpty(path) ){
      return filter != null ?
        filter( PathMap.get<T>(path) ?? defaultValue ) :
        ( PathMap.get<T>(path) ?? defaultValue );
    }
    if( listen ) this._addPath(path);
    return filter != null ?
      filter( PathMap.get<T>(path) ?? defaultValue ) :
      ( PathMap.get<T>(path) ?? defaultValue );
  }
  /// Get text from pathmap in conjunction with UIWidget.
  /// 
  /// Displays anything as long as Path is Unit and can be converted to text.
  /// 
  /// If not, [defaultValue] will be displayed.
  /// 
  /// [path]: The path to get.
  /// [defaultValue]: Value if none.
  /// [filter]: Filter values.
  /// [listen]: True to listen for values.
  String text(
    String path,
    {
      String defaultValue = Const.empty,
      String filter( String value ),
      bool listen = true
    }
  ) => filter != null ?
    filter( this.get<dynamic>( path, listen: listen )?.toString() ?? defaultValue ) :
    ( this.get<dynamic>( path, listen: listen )?.toString() ?? defaultValue );
  /// Get localized text from pathmap in conjunction with UIWidget.
  /// 
  /// Displays anything as long as Path is Unit and can be converted to text.
  /// 
  /// If not, [defaultValue] will be displayed.
  /// 
  /// [path]: The path to get.
  /// [defaultValue]: Value if none.
  /// [filter]: Filter values.
  /// [listen]: True to listen for values.
  String localized(
    String path,
    {
      String defaultValue = Const.empty,
      String filter( String value ), 
      bool listen = true
    }
  ) => filter != null ?
    filter( Localize.get( this.get<dynamic>( path, listen: listen )?.toString() ?? defaultValue ) ) :
    ( Localize.get( this.get<dynamic>( path, listen: listen )?.toString() ?? defaultValue ) );
  /// Get number from pathmap in conjunction with UIWidget.
  /// 
  /// Displays anything as long as Path is Unit and can be converted to double.
  /// 
  /// If not, [defaultValue] will be displayed.
  /// 
  /// [path]: The path to get.
  /// [defaultValue]: Value if none.
  /// [filter]: Filter values.
  /// [listen]: True to listen for values.
  double number(
    String path,
    {
      double defaultValue = 0,
      double filter( double value ),
      bool listen = true
    }
  ) {
    dynamic tmp = this.get<dynamic>( path, listen: listen );
    if( tmp is double ) return filter != null ? filter( tmp ) : tmp;
    if( tmp is int ) return filter != null ? filter( tmp as double ) : tmp as int;
    if( tmp is bool ) return filter != null ? filter( tmp ? 1 : 0 ) : ( tmp ? 1 : 0 );
    if( tmp is String ){
      return filter != null ?
      filter( double.tryParse( tmp ) ?? defaultValue ) :
      ( double.tryParse( tmp ) ?? defaultValue );
    }
    return filter != null ? filter( defaultValue ) : defaultValue;
  }
  /// Get boolean flag from pathmap in conjunction with UIWidget.
  /// 
  /// Displays anything as long as Path is Unit and can be converted to double.
  /// 
  /// If not, [defaultValue] will be displayed.
  /// 
  /// [path]: The path to get.
  /// [defaultValue]: Value if none.
  /// [filter]: Filter values.
  /// [listen]: True to listen for values.
  bool boolean(
    String path,
    {
      bool defaultValue = false,
      bool filter( bool value ),
      bool listen = true
    }
  ) {
    dynamic tmp = this.get<dynamic>( path, listen: listen );
    if( tmp is bool ) return filter != null ? filter( tmp ) : tmp;
    if( tmp is double ) return filter != null ? filter( tmp != 0 ) : tmp != 0;
    if( tmp is int ) return filter != null ? filter( tmp != 0 ) : tmp != 0;
    if( tmp is String ){
      return filter != null ?
        filter( tmp.toLowerCase() == "true" ) :
        ( tmp.toLowerCase() == "true" );
    }
    return filter != null ? filter( defaultValue ) : defaultValue;
  }
  /// Get the action callback from pathmap in conjunction with UIWidget.
  /// 
  /// If it does not exist, the action that does nothing is called.
  /// 
  /// [path]: The path to get.
  /// [listen]: True to listen for values.
  /// [defaultAction]: Default callback when no value exists in path.
  VoidAction action(
    String path,
    {
      bool listen = false,
      VoidAction defaultAction
    }
  ){
    return this.get<VoidAction>( path, listen: listen ) ?? defaultAction ?? (){};
  }

  /// Gets the current context.
  BuildContext get context => this._container?.context;
  /// Gets the current widget.
  Widget get widget => this._state.widget;
  void _addPath( String path ){
    if( isEmpty(path) || this._data.containsKey(path) ) return;
    Observer observer = Observer();
    PathListener.listen(path, this._notifyUpdate, observer:observer );
    this._data[path] = observer;
    Log.ast(
      "Add rebuild listener: %s (%s) %d at %s",
      [
        this._state.widget.runtimeType,
        this._state.widget.key?.toString(),
        this._updatedTime,
        path
      ]
    );
  }
  void _notifyUpdate( dynamic object ) {
    if( object is IPath ){
      int updatedTime = object.asType<IPath>().updatedTime;
      if( this._updatedTime >= updatedTime ) return;
      this._updatedTime = updatedTime;
    }
    this._state?._notifyUpdate( object );
  }
  void _dispose(){
    for( Observer tmp in this._data.values ) tmp.dispose();
    this._willDisposePathList.dispose();
    this._data.release();
    this._state = null;
  }
}